#!/usr/bin/env python3
"""
Malware Hunter
Advanced malware detection and analysis tool.
Parameters:
    - scan_path: Directory to scan
    - mode: Scan mode (quick, deep, heuristic)
    - quarantine: Enable quarantine of suspicious files
Requires: admin/root privileges for system scanning
"""
import os
import hashlib
import logging
from datetime import datetime
import json
from pathlib import Path
import threading
import queue
import time
import re
import magic  # python-magic for file type detection
import yara  # yara-python for pattern matching
import psutil
import platform
import subprocess
import requests
import tempfile
import shutil
import zipfile
import argparse

class MalwareHunter:
    def __init__(self, scan_path, mode='quick', enable_quarantine=True):
        self.scan_path = Path(scan_path)
        self.mode = mode
        self.enable_quarantine = enable_quarantine
        
        self.quarantine_dir = Path("quarantine")
        self.quarantine_dir.mkdir(exist_ok=True)
        
        self.file_queue = queue.Queue()
        self.result_queue = queue.Queue()
        self.stop_event = threading.Event()
        
        # Initialize detection components
        self.setup_logging()
        self.load_yara_rules()
        self.initialize_signatures()
        
    def setup_logging(self):
        """Set up logging configuration"""
        log_dir = Path("malware_logs")
        log_dir.mkdir(exist_ok=True)
        
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(levelname)s - %(message)s',
            handlers=[
                logging.FileHandler(log_dir / "malware_hunter.log"),
                logging.StreamHandler()
            ]
        )
        
    def load_yara_rules(self):
        """Load YARA rules for malware detection"""
        rules_dir = Path("yara_rules")
        rules_dir.mkdir(exist_ok=True)
        
        # Create default rules if none exist
        if not list(rules_dir.glob("*.yar")):
            default_rules = """
rule Suspicious_Strings {
    strings:
        $cmd1 = "cmd.exe" nocase
        $cmd2 = "powershell" nocase
        $net1 = "download" nocase
        $net2 = "upload" nocase
        $enc1 = "base64" nocase
        $enc2 = "encrypt" nocase
        $sys1 = "System32" nocase
        $sys2 = "regedit" nocase
    condition:
        2 of them
}

rule Malicious_Behavior {
    strings:
        $proc1 = "CreateRemoteThread" nocase
        $proc2 = "VirtualAlloc" nocase
        $proc3 = "WriteProcessMemory" nocase
        $net1 = "WSASocket" nocase
        $net2 = "connect" nocase
        $file1 = "CreateFile" nocase
        $file2 = "WriteFile" nocase
    condition:
        3 of them
}

rule Ransomware_Indicators {
    strings:
        $ext1 = ".encrypted" nocase
        $ext2 = ".locked" nocase
        $msg1 = "your files" nocase
        $msg2 = "bitcoin" nocase
        $msg3 = "payment" nocase
    condition:
        2 of them
}
"""
            with open(rules_dir / "default_rules.yar", "w") as f:
                f.write(default_rules)
                
        # Compile YARA rules
        rules_files = list(rules_dir.glob("*.yar"))
        self.yara_rules = yara.compile(filepaths={
            f.stem: str(f) for f in rules_files
        })
        
    def initialize_signatures(self):
        """Initialize malware signatures and heuristics"""
        self.signatures = {
            'file_types': [
                'application/x-dosexec',
                'application/x-executable',
                'application/x-sharedlib'
            ],
            'extensions': [
                '.exe', '.dll', '.sys',
                '.vbs', '.ps1', '.bat',
                '.tmp', '.scr'
            ],
            'patterns': [
                rb'MZ[\x90\x00]',  # PE file header
                rb'TVqQAAMAAAAEAAAA',  # Base64 encoded MZ header
                rb'powershell.*-enc',
                rb'cmd\.exe.*\/c'
            ]
        }
        
    def calculate_file_hash(self, filepath):
        """Calculate multiple hashes of file"""
        try:
            md5 = hashlib.md5()
            sha1 = hashlib.sha1()
            sha256 = hashlib.sha256()
            
            with open(filepath, "rb") as f:
                for chunk in iter(lambda: f.read(4096), b""):
                    md5.update(chunk)
                    sha1.update(chunk)
                    sha256.update(chunk)
                    
            return {
                'md5': md5.hexdigest(),
                'sha1': sha1.hexdigest(),
                'sha256': sha256.hexdigest()
            }
        except Exception as e:
            logging.error(f"Error calculating hash for {filepath}: {str(e)}")
            return None
            
    def check_virustotal(self, file_hash):
        """Check file hash against VirusTotal (requires API key)"""
        api_key = os.getenv('VIRUSTOTAL_API_KEY')
        if not api_key:
            return None
            
        try:
            url = f"https://www.virustotal.com/vtapi/v2/file/report"
            params = {
                'apikey': api_key,
                'resource': file_hash
            }
            response = requests.get(url, params=params)
            return response.json() if response.status_code == 200 else None
        except:
            return None
            
    def analyze_pe_file(self, filepath):
        """Analyze PE file characteristics"""
        try:
            import pefile
            pe = pefile.PE(filepath)
            
            characteristics = {
                'is_dll': pe.is_dll(),
                'is_exe': pe.is_exe(),
                'imports': [],
                'sections': []
            }
            
            # Get imports
            if hasattr(pe, 'DIRECTORY_ENTRY_IMPORT'):
                for entry in pe.DIRECTORY_ENTRY_IMPORT:
                    for imp in entry.imports:
                        characteristics['imports'].append({
                            'dll': entry.dll.decode(),
                            'name': imp.name.decode() if imp.name else ''
                        })
                        
            # Get sections
            for section in pe.sections:
                characteristics['sections'].append({
                    'name': section.Name.decode().rstrip('\x00'),
                    'size': section.SizeOfRawData,
                    'entropy': section.get_entropy()
                })
                
            return characteristics
        except:
            return None
            
    def scan_file(self, filepath):
        """Scan individual file for malware"""
        try:
            result = {
                'file': str(filepath),
                'size': os.path.getsize(filepath),
                'timestamp': datetime.now().isoformat(),
                'threats': []
            }
            
            # Calculate file hashes
            hashes = self.calculate_file_hash(filepath)
            if hashes:
                result.update(hashes)
                
            # Check file type
            file_type = magic.from_file(str(filepath), mime=True)
            result['type'] = file_type
            
            # Basic checks
            if any(filepath.suffix.lower() == ext for ext in self.signatures['extensions']):
                result['threats'].append({
                    'type': 'suspicious_extension',
                    'detail': filepath.suffix
                })
                
            if file_type in self.signatures['file_types']:
                result['threats'].append({
                    'type': 'suspicious_filetype',
                    'detail': file_type
                })
                
            # Content analysis
            try:
                with open(filepath, 'rb') as f:
                    content = f.read()
                    
                    # Pattern matching
                    for pattern in self.signatures['patterns']:
                        if re.search(pattern, content):
                            result['threats'].append({
                                'type': 'suspicious_pattern',
                                'detail': pattern.decode()
                            })
                            
                    # YARA rules matching
                    matches = self.yara_rules.match(data=content)
                    if matches:
                        result['threats'].append({
                            'type': 'yara_match',
                            'detail': [m.rule for m in matches]
                        })
                        
            except Exception as e:
                logging.error(f"Error analyzing file content {filepath}: {str(e)}")
                
            # Deep analysis for executables
            if self.mode in ['deep', 'heuristic'] and file_type == 'application/x-dosexec':
                pe_analysis = self.analyze_pe_file(filepath)
                if pe_analysis:
                    result['pe_analysis'] = pe_analysis
                    
                    # Check for suspicious imports
                    suspicious_imports = [
                        'VirtualAlloc',
                        'WriteProcessMemory',
                        'CreateRemoteThread'
                    ]
                    
                    for imp in pe_analysis['imports']:
                        if any(sus in imp['name'] for sus in suspicious_imports):
                            result['threats'].append({
                                'type': 'suspicious_import',
                                'detail': imp['name']
                            })
                            
                    # Check section entropy
                    for section in pe_analysis['sections']:
                        if section['entropy'] > 7.0:  # High entropy indicates encryption/packing
                            result['threats'].append({
                                'type': 'high_entropy',
                                'detail': f"Section {section['name']}: {section['entropy']}"
                            })
                            
            # VirusTotal check in heuristic mode
            if self.mode == 'heuristic' and result['threats']:
                vt_result = self.check_virustotal(result['sha256'])
                if vt_result:
                    result['virustotal'] = vt_result
                    
            return result
            
        except Exception as e:
            logging.error(f"Error scanning file {filepath}: {str(e)}")
            return None
            
    def quarantine_file(self, filepath, scan_result):
        """Move suspicious file to quarantine"""
        try:
            if not self.enable_quarantine:
                return
                
            # Create quarantine subdirectory based on threat level
            threat_level = 'high' if len(scan_result['threats']) > 2 else 'medium'
            quarantine_subdir = self.quarantine_dir / threat_level
            quarantine_subdir.mkdir(exist_ok=True)
            
            # Create encrypted zip with original file
            quarantine_file = quarantine_subdir / f"{filepath.name}.encrypted"
            
            with zipfile.ZipFile(quarantine_file, 'w', zipfile.ZIP_DEFLATED) as zipf:
                zipf.write(filepath, filepath.name, zipfile.ZIP_DEFLATED)
                
            # Store metadata
            metadata = {
                'original_path': str(filepath),
                'quarantine_time': datetime.now().isoformat(),
                'scan_result': scan_result
            }
            
            with open(quarantine_file.with_suffix('.meta'), 'w') as f:
                json.dump(metadata, f, indent=2)
                
            # Remove original file
            os.remove(filepath)
            logging.info(f"Quarantined file: {filepath}")
            
            return str(quarantine_file)
            
        except Exception as e:
            logging.error(f"Error quarantining file {filepath}: {str(e)}")
            return None
            
    def scan_directory(self):
        """Scan directory for malware"""
        logging.info(f"Starting malware scan of {self.scan_path}")
        logging.info(f"Mode: {self.mode}")
        
        start_time = time.time()
        results = []
        
        try:
            # Walk through directory
            for root, _, files in os.walk(self.scan_path):
                for file in files:
                    if self.stop_event.is_set():
                        break
                        
                    filepath = Path(root) / file
                    
                    # Skip if file is too large
                    if os.path.getsize(filepath) > 100 * 1024 * 1024:  # 100MB limit
                        logging.warning(f"Skipping large file: {filepath}")
                        continue
                        
                    # Scan file
                    result = self.scan_file(filepath)
                    if result:
                        results.append(result)
                        
                        # Quarantine if threats found
                        if result['threats'] and self.enable_quarantine:
                            quarantine_path = self.quarantine_file(filepath, result)
                            if quarantine_path:
                                result['quarantined'] = quarantine_path
                                
            # Save final results
            self.save_results(results, time.time() - start_time)
            
        except Exception as e:
            logging.error(f"Error during directory scan: {str(e)}")
            
    def save_results(self, results, duration):
        """Save scan results"""
        output = {
            'timestamp': datetime.now().isoformat(),
            'scan_path': str(self.scan_path),
            'mode': self.mode,
            'duration': duration,
            'total_files': len(results),
            'threats_found': sum(1 for r in results if r['threats']),
            'results': results
        }
        
        # Save to file
        output_dir = Path("malware_logs")
        output_file = output_dir / f"scan_{datetime.now().strftime('%Y%m%d_%H%M%S')}.json"
        with open(output_file, 'w') as f:
            json.dump(output, f, indent=2)
            
        logging.info(f"Scan results saved to {output_file}")
        
        # Print summary
        print("\nMalware Scan Summary")
        print("=" * 50)
        print(f"Scan Path: {self.scan_path}")
        print(f"Mode: {self.mode}")
        print(f"Duration: {duration:.1f} seconds")
        print(f"Total Files Scanned: {output['total_files']}")
        print(f"Threats Found: {output['threats_found']}")
        
        if output['threats_found'] > 0:
            print("\nDetected Threats:")
            for result in results:
                if result['threats']:
                    print(f"\nFile: {result['file']}")
                    for threat in result['threats']:
                        print(f"- {threat['type']}: {threat['detail']}")
                    if 'quarantined' in result:
                        print(f"* Quarantined to: {result['quarantined']}")

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Malware Hunter")
    parser.add_argument("path", help="Directory to scan")
    parser.add_argument("--mode", choices=['quick', 'deep', 'heuristic'],
                      default='quick', help="Scan mode")
    parser.add_argument("--no-quarantine", action="store_true",
                      help="Disable quarantine")
    args = parser.parse_args()
    
    hunter = MalwareHunter(args.path, args.mode, not args.no_quarantine)
    hunter.scan_directory()
